import sys
from unittest import TestCase, main as unittest_main
import moordyn


class GeneratorTests(TestCase):
    def setUp(self):
        pass

    def test_generation(self):
        system = moordyn.Generator.Mooring()
        line_type = moordyn.Generator.LineMaterial(
            "main", 0.09, 77.7066, 384.243E6, -0.8, 0, 1.6, 1.0, 0.1, 0.0)
        system.AddLineMaterial(line_type)

        points = [
            moordyn.Generator.Point(
                "Fixed", [853.87, 0, -320.0], 0, 0, 0, 0),
            moordyn.Generator.Point(
                "Fixed", [-426.94, 739.47, -320.0], 0, 0, 0, 0),
            moordyn.Generator.Point(
                "Fixed", [-426.94, -739.47, -320.0], 0, 0, 0, 0),
            moordyn.Generator.Point(
                "Coupled", [5.2, 0, -70.0], 0, 0, 0, 0),
            moordyn.Generator.Point(
                "Coupled", [-2.6, 4.5, -70.0], 0, 0, 0, 0),
            moordyn.Generator.Point(
                "Coupled", [-2.6, -4.5, -70.0], 0, 0, 0, 0),
        ]
        for point in points:
            system.AddPoint(point)

        lines = []
        for i in range(3):
            lines.append(moordyn.Generator.Line(
                line_type,
                moordyn.Generator.LinePoint(points[i]),
                moordyn.Generator.LinePoint(points[i + 3]),
                902.2,
                20
            ))
        for line in lines:
            system.AddLine(line)

        system.AddOption(moordyn.Generator.Option("writeLog", 2))
        system.AddOption(moordyn.Generator.Option("dtM", 0.002))
        system.AddOption(moordyn.Generator.Option("kBot", 3.0e6))
        system.AddOption(moordyn.Generator.Option("cBot", 3.0e5))
        system.AddOption(moordyn.Generator.Option("WtrDnsty", 1025.0))
        system.AddOption(moordyn.Generator.Option("WtrDpth", 320))
        system.AddOption(moordyn.Generator.Option("dtIC", 1.0))
        system.AddOption(moordyn.Generator.Option("TmaxIC", 100.0))
        system.AddOption(moordyn.Generator.Option("CdScaleIC", 4.0))
        system.AddOption(moordyn.Generator.Option("threshIC", 0.001))

        system = system.Create()

        x = []
        for i in range(4, 7):
            point = moordyn.GetPoint(system, i)
            x = x + list(moordyn.GetPointPos(point))
        v = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(moordyn.Init(system, x, v), 0,
                         "Failure initializing the lines")
        self.assertEqual(moordyn.NCoupledDOF(system), 9,
                         "Wrong number of coupled DOFs")
        self.assertEqual(moordyn.GetNumberLines(system), 3)
        v[0] = 0.1
        forces = moordyn.Step(system, x, v, 0.0, 0.5)
        self.assertEqual(moordyn.Close(system),
                         0, "Failure finishing MoorDyn")


if __name__ == '__main__':
    unittest_main()



from itertools import product
import os
from dataclasses import dataclass

"""Simple script for generating a bunch of model files for benchmarking purposes.

Just needs basic python (tested for version >=3.8).
"""


@dataclass
class BenchmarkParams:
    num_lines: int
    num_segments: int


def generate_benchmark(base_output_folder, params: BenchmarkParams):
    """Generate a single benchmark model file.

    The model has a user defined number of lines that each share a fixed top
    point, and have their own bottom point.
    They start off at a 10 degree angle from vertical.
    """
    points = "\n".join(
        f"{i + 2} free 34.20201433256687 0.0 -93.96926207859084 0 0 0 0 -"
        for i in range(params.num_lines)
    )
    line_length = 100.0
    lines = "\n".join(
        f"{i+1} chain 1 {i + 2} {line_length} {params.num_segments} p,t"
        for i in range(params.num_lines)
    )

    file_str = f"""--------------------- MoorDyn Input File -------------------------------------------------------
Output from generate_cases.py
----------------------- LINE TYPES --------------------------------------------------------------
TypeName   Diam    Mass/m     EA         BA/-zeta    EI         Cd     Ca     CdAx    CaAx
(name)     (m)     (kg/m)     (N)        (N-s/-)     (N-m^2)    (-)    (-)    (-)     (-)
chain      0.252   390        16000000.0 -0.3        0.0        1.37   0.64   1.0     0.0
----------------------- POINTS -------------------------------------------------------------------
Node      Type      X        Y         Z        M        V         CdA    CA
(-)       (-)       (m)      (m)       (m)      (kg)     (m^3)     (m^2)  (-)
1         free      0.0      0.0       0.0      0        0         0      0 -
{points}
-------------------------- LINES -----------------------------------------------------------------
Line     LineType NodeA     NodeB  UnstrLen  NumSegs     Flags/Outputs
(-)      (-)       (-)       (-)   (m)         (-)          (-)
{lines}
-------------------------- SOLVER OPTIONS---------------------------
0        writeLog     - Write a log file
0.001 dtM          - time step to use in mooring integration
rk4      tScheme
3000000.0    kb           - bottom stiffness
300000.0    cb           - bottom damping
200.0    WtrDpth      - water depth
4.0      ICDfac       - factor to scale drag coeff during IC
1.5e-06    threshIC     - threshold for IC convergence
0.0    TmaxIC       - threshold for IC convergence
1e-05     dtIC        - Time lapse between convergence tests (s)
9.81     g     - gravitational force constant
1025     rho     - water density
0     WaveKin     - waves mode
0     Currents     - current mode
------------------------- need this line --------------------------------------
"""

    folder_name = f"{params.num_lines}_lines_{params.num_segments}_segs"

    output_folder = os.path.join(base_output_folder, folder_name)
    try:
        os.mkdir(output_folder)
    except FileExistsError:
        pass

    out_path = os.path.join(output_folder, "lines.txt")

    with open(out_path, "w") as f:
        f.write(file_str)


def main():

    number_of_lines = [2**n for n in range(7)]
    number_of_segments = [2**n for n in range(7)]

    for vals in product(number_of_lines, number_of_segments):
        params = BenchmarkParams(*vals)
        generate_benchmark("./", params)


if __name__ == "__main__":
    main()



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# http://www.sphinx-doc.org/en/master/config

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import subprocess
import shutil
import os
import sys
sys.path.insert(0, os.path.abspath('../wrappers/python/moordyn/'))


# -- Doxygen and breathe steps -----------------------------------------------------------

# Check if we're running on Read the Docs' servers
read_the_docs_build = os.environ.get('READTHEDOCS', None) == 'True'

breathe_projects = {}

os.makedirs("_build/doxygen", exist_ok=True)
if not read_the_docs_build:
    subprocess.call('make clean', shell=True)
subprocess.call('doxygen', shell=True)
os.makedirs("_build/doxygen/out/doxygen", exist_ok=True)
if os.path.exists('_build/doxygen/out/doxygen/html'):
    shutil.rmtree('_build/doxygen/out/doxygen/html')
shutil.move('_build/doxygen/html', '_build/doxygen/out/doxygen/html')

breathe_projects['MoorDyn'] = "_build/doxygen/xml"
breathe_default_project = "MoorDyn"

# breathe_projects = { "trianglelib": "../../doxygen/build/xml/" }
# breathe_default_project = "trianglelib"


# -- Project information -----------------------------------------------------

project = 'MoorDyn'
copyright = '2024, National Renewable Energy Laboratory'
author = 'Matt Hall'

# The full version, including alpha/beta/rc tags
release = '2.3.1' # TODO: can we automate this?


# -- General configuration ---------------------------------------------------

master_doc = 'index'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
# extensions = ['recommonmark']
# extensions = ['sphinx.ext.autodoc', 'sphinx.ext.napoleon']
extensions = ["breathe", 'sphinx.ext.autodoc',
              'sphinx.ext.napoleon', 'sphinx.ext.mathjax']

napoleon_include_init_with_doc = True

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
# need to run pip install sphinx_rtd_theme to use locally
html_theme = 'sphinx_rtd_theme'
html_theme_options = {
    'collapse_navigation': False
}

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ['_static']

# specify custom attributes to accept (see https://github.com/sphinx-doc/sphinx/issues/2682)
cpp_id_attributes = ['DECLDIR']

# Keep the doxygen documentaion in readthedocs
html_extra_path = ['_build/doxygen/out']



"""
Copyright (c) 2024, Jose Luis Cercos-Pita <jlc@core-marine.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

def moorpy_export(moorpy_system, filename,
                  version_major=2, version_minor=4):
    """Generate an Initial Condition from a MoorPy already generated and solved
    system

    Parameters
    ----------
    moorpy_system (moorpy.System): The MoorPy instance
    filename (str): The output file path
    version_major (int): MoorDyn major version
    version_minor (int): MoorDyn minor version
    """
    import moorpy
    assert isinstance(moorpy_system, moorpy.System)
    import struct
    from scipy.spatial.transform import Rotation as R
    with open(filename, 'wb') as out:
        # Write the header
        out.write(b'MoorDyn')
        out.write(struct.pack('<B', version_major))
        out.write(struct.pack('<B', version_minor))
        data = b''
        ninstances = 0
        for body in moorpy_system.bodyList:
            ninstances += 1
            r = R.from_matrix(body.R)
            data += struct.pack('<Q', 1)
            data += struct.pack('<Q', 13)
            data += struct.pack('<ddd', *body.r6[:3])
            data += struct.pack('<dddd', *r.as_quat())
            data += struct.pack('<d', 0.0) * 6
        for rod in moorpy_system.rodList:
            ninstances += 1
            if isinstance(rod, moorpy.Line):
                x, y, z, _ = rod.getLineCoords(-1)
                rA = np.asarray([x[0], y[0], z[0]])
                rB = np.asarray([x[-1], y[-1], z[-1]])
                L = rod.L
            else:
                rA = rB = rod.r
                L = 0
            if L > 0.0:
                k = (rB - rA) / L
                Rmat = np.array(rotationMatrix(
                    0,
                    np.arctan2(np.hypot(k[0],k[1]), k[2]),
                    np.arctan2(k[1],k[0])))
                r = R.from_matrix(Rmat)
            else:
                r = R.from_matrix(np.eye(3))
            data += struct.pack('<Q', 1)
            data += struct.pack('<Q', 13)
            data += struct.pack('<ddd', *rA)
            data += struct.pack('<dddd', *r.as_quat())
            data += struct.pack('<d', 0.0) * 6
        for point in moorpy_system.pointList:
            ninstances += 1
            data += struct.pack('<Q', 1)
            data += struct.pack('<Q', 6)
            data += struct.pack('<ddd', *point.r)
            data += struct.pack('<d', 0.0) * 3
        for line in moorpy_system.lineList:
            x, y, z, _ = line.getLineCoords(-1)
            n = len(x)
            ninstances += 1
            data += struct.pack('<Q', n - 2)
            data += struct.pack('<Q', 6)
            for i in range(1, n - 1):
                data += struct.pack('<ddd', x[i], y[i], z[i])
                data += struct.pack('<d', 0.0) * 3
        # Save it
        out.write(struct.pack('<Q', len(data) // 8 + 1))
        out.write(struct.pack('<Q', ninstances))
        out.write(data)


def moorpy_ic(infile,
              outfile=None, tol=0.05, version_major=2, version_minor=4):
    """Generate an Initial Condition from a MoorPy/Moordyn input file. If
    MoorPy fails to solve the equilibrium, no Initial Condition file will be
    printed

    Parameters
    ----------
    infile (str): The output file path

    Keyword arguments
    -----------------
    outfile (str): The output file path. If None, infile with a suffix ".ic"
                   will be considered
    tol (float): The absolute tolerance on positions when calculating
                 equilibrium [m]
    version_major (int): MoorDyn major version
    version_minor (int): MoorDyn minor version

    Returns
    -------
    success (bool): True/False whether converged to within tolerance.
    """
    import moorpy
    print("WARNING: MoorPy does not support all features of MoorDyn, some states may not initialize")
    outfile = outfile or infile + ".ic"
    system = moorpy.System(file=infile)
    system.initialize()
    try:
        success = system.solveEquilibrium(tol=tol)
    except moorpy.SolveError:
        success = False
    if success:
        moorpy_export(system, outfile)
    return success



from unittest import TestCase, main as unittest_main
import os
import tempfile
import moordyn


def setup_case(filename):
    """Setup a case in a temporal folder

    Parameters:
    filename (str): The filename in the "Mooring" subfolder

    Returns:
    str: The folder where the case has been generated
    """
    tmp_folder = tempfile.mkdtemp()
    # The input should be in the Mooring/lines.txt subfolder.
    # See MoorDyn.cpp:287
    input_folder = os.path.join(tmp_folder, "Mooring")
    os.mkdir(input_folder)
    r_path = os.path.join(
        os.path.dirname(__file__), "Mooring", filename)
    w_path = os.path.join(input_folder, "lines.txt")
    with open(r_path, "r") as r_file, open(w_path, "w") as w_file:
        w_file.write(r_file.read())
    return tmp_folder



class MinimalTests(TestCase):
    def setUp(self):
        pass

    def test_no_input(self):
        try:
            system = moordyn.Create('file.invalid')
        except RuntimeError:
            return
        self.assertEqual(system, None, "The creation should fail")

    def test_one_iter(self):
        tmp_folder = setup_case("lines.txt")
        cwd = os.getcwd()
        os.chdir(tmp_folder)
        system = moordyn.Create()
        os.chdir(cwd)
        x = []
        for i in range(4, 7):
            point = moordyn.GetPoint(system, i)
            x = x + list(moordyn.GetPointPos(point))
        v = [0, 0, 0, 0, 0, 0, 0, 0, 0]
        self.assertEqual(moordyn.Init(system, x, v), 0,
                         "Failure initializing the lines")
        self.assertEqual(moordyn.NCoupledDOF(system), 9,
                         "Wrong number of coupled DOFs")
        self.assertEqual(moordyn.GetNumberLines(system), 3)
        v[0] = 0.1
        forces = moordyn.Step(system, x, v, 0.0, 0.5)
        print(forces)
        self.assertEqual(moordyn.Close(system),
                         0, "Failure finishing MoorDyn")
        f_ref = 7.2e5
        self.assertTrue(abs(forces[0] - f_ref) / f_ref < 0.01,
                        "Wrong surge force")


if __name__ == '__main__':
    unittest_main()



import numpy as np
import matplotlib.pyplot as plt


START_Y = 0.0
TARGET_Y = 1.0
C_0 = 0.0
C_1 = 0.075
OMEGA = 2. * np.pi
DT_INC = 0.1
DDRDDT_TOL = 1e-6


def relax_constant(i, n):
    return C_0 / n


def relax_tanh(i, n):
    x = (i + 1) / n
    return C_1 * (1 / n) * np.tanh(x)


def relax_bell(i, n):
    x = (i + 1) / n
    if x > 0.5:
        x = 1. - x
    return C_1 * (1 / n) * np.tanh(x)


def relax_all(i, n):
    x = (i + 1) / n
    return relax_constant(i, n) + relax_tanh(i, n)


def relax(y0, y1, i, n, backend=relax_all):
    f = backend(i, n)
    return (1. - f) * y0 + f * y1


def oraculus(iters, backend=relax_all):
    x = np.arange(0, iters + 1)
    y = [START_Y, ]
    f = [0.0, ]
    for i in range(iters - 1):
        y.append(relax(y[-1], TARGET_Y, i, iters, backend=backend))
        f.append(backend(i, iters))
    y.append(TARGET_Y)
    f.append(1.0)
    return x, np.asarray(y), np.asarray(f)


def oraculus_plot(N=(10, 100, 1000)):
    fig, (ax1, ax2) = plt.subplots(2, 1)
    for n, c in zip(N, ('k', 'r', 'b')):
        x, y, f = oraculus(n)
        ax1.plot(x / n, y, color=c,
                 label=f'M = {n}, C_0 = {C_0}, C_1 = {C_1}')
        ax2.plot(x[1:-1] / n, 1 - f[1:-1], color=c,
                    label=f'M = {n}, C_0 = {C_0}, C_1 = {C_1}')

    ax1.set_ylabel(r'$\mathrm{d} r / \mathrm{d} t$')
    ax1.set_xlim(0.0, 1.0)
    ax1.set_ylim(0.0, 1.0)
    ax2.set_xlabel(r'$m / M$')
    ax2.set_ylabel('$1 - f$')
    ax2.set_xlim(0.0, 1.0)
    ax1.grid()
    ax2.grid()
    ax1.legend(loc='best')
    plt.show()
    

def ode2(iters, r0=1, backend=relax_all):
    dt_0 = dt = 2. / OMEGA * iters
    while True:
        r = r0
        ddrddt = -OMEGA**2 * r
        sol = -OMEGA**2 * r / (1. + 0.5 * OMEGA**2 * dt**2)
        for i in range(iters):
            r = r0 + 0.5 * dt**2 * ddrddt
            ddrddt_new = -OMEGA**2 * r
            if i < iters - 1:
                ddrddt = relax(ddrddt, ddrddt_new, i, iters, backend=backend)
            else:
                ddrddt = ddrddt_new
            if np.abs(ddrddt - sol) < DDRDDT_TOL:
                break
        if ddrddt > 0 or -0.5 * dt**2 * ddrddt > 2. * r0:
            return dt - DT_INC * dt_0
        dt += DT_INC * dt_0
    return dt


def ode2_plot(N=range(5, 101, 2)):
    global C_0, C_1
    c0, c1 = C_0, C_1
    dts = []
    for n in N:
        C_0 = max(np.polyval([-0.01, 0.1], [n])[0], 0)
        if n >= 10:
            C_1 = max(1 / np.polyval([0.051, 0, 10.0], [n])[0], 0)
        else:
            C_1 = 0.07
        dt = ode2(n)
        dts.append(dt)
    dts = np.asarray(dts) * OMEGA / 2 / N
    plt.plot(N, dts, 'k-')
    plt.plot([N[0], N[-1]], [1.0, 1.0], 'k--')
    plt.xlabel(r'$M$')
    plt.ylabel(r'$\frac{\Delta t \omega}{2 N}$')
    plt.grid()
    # for x, y, c0p, c1p in zip(N, dts, c0s, c1s):
    #     plt.text(x, y, f'{c0p:.3f},{c1p:.3f}')
    plt.show()
    C_0, C_1 = c0, c1


if __name__ == '__main__':
    # oraculus_plot()
    ode2_plot()



import numpy as np

T = 15.0
A = 1.0
UX = 1.25
x = np.linspace(-50.0, 50.0, num=11)
y = np.linspace(-2.0, 2.0, num=2)
z = np.linspace(-30.0, 5.0, num=36)
# t = np.linspace(0.0, 15.0, num=101)
t = np.linspace(0.0, 15.0, num=2)
zeta = A * np.sin(2.0 * np.pi * t / T)

# Define a current that increases in X dimension, pointing along the positive x-axis
# Doesn't vary current in any other dimension.
xx, yy, zz, tt = np.meshgrid(x, y, z, t)
ux = 0.1 * xx + 0.1 * t
uy = 0.0 * yy
uz = 0.05 * zz

with open('water_grid.txt', 'w') as f:
    f.write('--------------------- MoorDyn Waves grid File ----------------------------------\n')
    f.write('List of grid points, in 3 blocks (x, y, z)\n')
    f.write(
        'Each block starts with a 2 (i.e. list of coords), and then th coordinates (m)\n')
    for l in (x, y, z):
        txt = '1\n'
        for p in l:
            txt = txt + str(p) + ' '
        f.write(txt.strip() + '\n')

with open('current_profile_4d.txt', 'w') as f:
    # First line is number of (x,y,z,t) coords
    f.write(f"{len(x)} {len(y)} {len(z)} {len(t)}\n")

    # Next four lines specify the axis (tick) values for x, y, z, t respectively
    for xx in x:
        f.write(f"{xx} ")
    f.write("\n")

    for yy in y:
        f.write(f"{yy} ")
    f.write("\n")

    for zz in z:
        f.write(f"{zz} ")
    f.write("\n")

    for tt in t:
        f.write(f"{tt} ")
    f.write("\n")

    # All the remaining lines specify an (x, y, z, t) coordinate followed by
    # a (ux, uy, uz) current vector at that coordinate
    for idx, xs in enumerate(x):
        for idy, ys in enumerate(y):
            for idz, zs in enumerate(z):
                for idt, ts in enumerate(t):
                    f.write(
                        f"{xs} {ys} {zs} {ts} {ux[idy][idx][idz][idt]:.6g} {uy[idy][idx][idz][idt]} {uy[idy][idx][idz][idt]}\n")



import numpy as np
BASE_DEPTH = -30
x = np.linspace(-50.0, 50.0, num=100)
y = np.linspace(-2.0, 2.0, num=5)

# Define 2d grid for seafloor:
xx, yy = np.meshgrid(x,y)

with open('seafloor_profile_3d.txt', 'w') as f:
    # First line is number of (x,y) coords
    f.write(f"{len(x)} {len(y)}\n")
    
    # Next two lines specify the axis (tick) values for x, y respectively
    for xx in x:
        f.write(f"{xx} ")
    f.write("\n")

    for yy in y:
        f.write(f"{yy} ")
    f.write("\n")

    # All the remaining lines specify an (x, y, z) coordinate
    for idx, xs in enumerate(x):
        for idy, ys in enumerate(y):
            f.write(f"{xs} {ys} {BASE_DEPTH + 0.1 * xs}\n")



import numpy as np
import matplotlib.pyplot as plt


START_Y = 0.0
TARGET_Y = 1.0
C_0 = 0.0
C_1 = 0.075
OMEGA = 2. * np.pi
DT_INC = 0.1
DDRDDT_TOL = 1e-3


def relax_constant(i, n):
    return C_0 / n


def relax_tanh(i, n):
    x = (i + 1) / n
    return C_1 * (1 / n) * np.tanh(x)


def relax_bell(i, n):
    x = (i + 1) / n
    if x > 0.5:
        x = 1. - x
    return C_1 * (1 / n) * np.tanh(x)


def relax_all(i, n):
    x = (i + 1) / n
    return relax_constant(i, n) + relax_tanh(i, n)


def relax(y0, y1, i, n, backend=relax_all):
    f = backend(i, n)
    return (1. - f) * y0 + f * y1


def ode2(iters, r0=1, backend=relax_all):
    dt_0 = dt = 2. / OMEGA * iters
    while True:
        r = r0
        ddrddt_0 = ddrddt = -OMEGA**2 * r0
        sol = -OMEGA**2 * (1. - 0.25 * OMEGA**2 * dt**2) * r0 / (1. + 0.25 * OMEGA**2 * dt**2)
        for i in range(iters):
            r = r0 + 0.25 * dt**2 * (ddrddt_0 + ddrddt)
            ddrddt_new = -OMEGA**2 * r
            if i < iters - 1:
                ddrddt = relax(ddrddt, ddrddt_new, i, iters, backend=backend)
            else:
                ddrddt = ddrddt_new
        if np.abs(ddrddt - sol) > DDRDDT_TOL:
            return dt - DT_INC * dt_0
        dt += DT_INC * dt_0
    return dt


def ode2_plot(N=range(2, 50, 1)):
    global C_0, C_1
    c0, c1 = C_0, C_1
    dts, c0s, c1s = [], [], []
    for n in N:
        print(n)
        best_dt, best_c0, best_c1 = 0.0, 0.0, 0.0
        # for C_0 in np.arange(0, 0.1, 0.001):
        #     for C_1 in np.arange(0.01, 0.5, 0.01):
        #         dt = ode2(n)
        #         if dt > best_dt:
        #             best_dt = dt
        #             best_c0 = C_0
        #             best_c1 = C_1
        # dts.append(best_dt)
        # c0s.append(best_c0)
        # c1s.append(best_c1)
        if n >= 14:
            C_0 = 0.01
            C_1 = 0.02
        else:
            C_0 = max(np.polyval([-0.011, 0.154], [n])[0], 0)
            C_1 = 0.08
        dt = ode2(n)
        dts.append(dt)
        c0s.append(C_0)
        c1s.append(C_1)
    dts = np.asarray(dts) * OMEGA / 2 / N
    plt.plot(N, dts, 'k-')
    plt.plot([N[0], N[-1]], [1.0, 1.0], 'k--')
    plt.xlabel(r'$M$')
    plt.ylabel(r'$\frac{\Delta t \omega}{2 N}$')
    plt.grid()
    plt.show()

    plt.plot(N, c0s, 'k-')
    plt.grid()
    plt.show()

    plt.plot(N, c1s, 'k-')
    plt.grid()
    plt.show()

    C_0, C_1 = c0, c1


if __name__ == '__main__':
    # oraculus_plot()
    ode2_plot()



import os
ROOT = os.path.dirname(os.path.realpath(__file__))
import sys
sys.path.append(os.path.join(ROOT, '../../wrappers/python/moordyn/'))
from moorpyic import *
import argparse


parser = argparse.ArgumentParser(
    prog='moorpy_ic',
    description='Create an Initial Condition using MoorPy')
parser.add_argument('filename')
parser.add_argument('-t', '--tol', type=float, default=0.05,
                    help='Equilibrium solver absolute tolerance')

with open(os.path.join(ROOT, '../../CMakeLists.txt'), 'r') as f:
    txt = f.read()
    def read_version_line(txt, name):
        subtxt = "set(" + name + " "
        txt = txt[txt.find(subtxt) + len(subtxt):]
        txt = txt[:txt.find(')\n')]
        return int(txt.strip())
    version_major = read_version_line(txt, "MOORDYN_MAJOR_VERSION")
    version_minor = read_version_line(txt, "MOORDYN_MINOR_VERSION")

args = parser.parse_args()
if not moorpy_ic(args.filename,
                 tol=args.tol,
                 version_major=version_major,
                 version_minor=version_minor):
    print("The system did not converge!")



from .Mooring import Mooring
from .Body import BodyPoint, Body
from .Point import Point
from .Failure import Failure
from .Line import LineMaterial, LinePoint, Line
from .Option import Option
from .Rod import RodMaterial, RodPoint, Rod



from unittest import TestCase, main as unittest_main
import os
import csv
import math
import moordyn
from test_minimal import setup_case


MBL = 25.3e6
KRS = 13.4
KRD1 = 16.0
KRD2 = 0.35
TC = 200.0
TTIMES = [320.0, 640.0, 960.0, 1280.0, 1599.0]
TMEANS = [0.1, 0.2, 0.3, 0.4, 0.5]


def vec_norm(v):
    return math.sqrt(v[0] * v[0] + v[1] * v[1] + v[2] * v[2])


def get_average_tension(line, anchor, fairlead):
    n = moordyn.GetLineN(line)
    # To average the tensions along the line, we are considering the
    # connections instead of the line end up nodes, as long as the latter are
    # not that accurate
    tension = 0.0
    # force = moordyn.GetPointForce(anchor)
    # tension += vec_norm(force)
    force = moordyn.GetPointForce(fairlead)
    tension += vec_norm(force)
    for i in range(n):
        force = moordyn.GetLineNodeTen(line, i)
        tension += vec_norm(force)
    return tension / (n + 1)


class PolyesterTests(TestCase):
    def setUp(self):
        pass

    def test_rampup_stabilization_cycles(self):
        global TTIMES, TMEANS

        tmp_folder = setup_case("polyester/simple.txt")
        cwd = os.getcwd()
        os.chdir(tmp_folder)
        system = moordyn.Create()
        os.chdir(cwd)

        n_dof = moordyn.NCoupledDOF(system)
        self.assertEqual(moordyn.NCoupledDOF(system), 3,
                         "Incorrect number of DOFs")

        anchor = moordyn.GetPoint(system, 1)
        fairlead = moordyn.GetPoint(system, 2)
        line = moordyn.GetLine(system, 1)

        l0 = moordyn.GetLineUnstretchedLength(line)
        r = moordyn.GetPointPos(fairlead)
        dr = [0, 0, 0]
        moordyn.Init(system, r, dr)

        tdata, xdata = [], []
        with open('Mooring/polyester/motion.csv') as csvfile:
            reader = csv.reader(csvfile, delimiter=',', quotechar='"')
            skip = 2
            for row in reader:
                if skip > 0:
                    skip -= 1
                    continue
                tdata.append(float(row[0]))
                xdata.append(float(row[1]))

        t = 0
        times, tensions = [], []
        for i in range(len(tdata)):
            t_dst = tdata[i]
            x_dst = xdata[i]
            dt = t_dst - t
            dr[0] = (x_dst - r[0]) / dt
            f = moordyn.Step(system, r, dr, t, dt)
            t += dt
            times.append(t)
            tensions.append(get_average_tension(line, anchor, fairlead))

            if (times[-1] - times[0] < TC):
                continue

            tension = 0.0;
            for f in tensions:
                tension += f
            tension /= len(tensions)
            times = times[1:]
            tensions = tensions[1:]

            ks = KRS * MBL
            kd = (KRD1 + KRD2 * tension / MBL * 100) * MBL
            l = l0 * (1.0 + tension / ks) / (1.0 + tension / kd)
            moordyn.SetLineConstantEA(line, kd)
            moordyn.SetLineUnstretchedLength(line, l)

            if (t >= TTIMES[0]):
                tmean = TMEANS[0]
                TTIMES = TTIMES[1:]
                TMEANS = TMEANS[1:]
                assert(abs(tension / MBL - tmean) < 0.025)

        moordyn.Close(system)

if __name__ == '__main__':
    unittest_main()



"""
Copyright (c) 2022-2024, Jose Luis Cercos-Pita <jlc@core-marine.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

LEVEL_NONE = 4096
LEVEL_ERR = 3
LEVEL_WRN = 2
LEVEL_MSG = 1
LEVEL_DBG = 0


ERRCODE_SUCCESS = 0
ERRCODE_INVALID_INPUT_FILE = -1
ERRCODE_INVALID_OUTPUT_FILE = -2
ERRCODE_INVALID_INPUT = -3
ERRCODE_NAN_ERROR = -4
ERRCODE_MEM_ERROR = -5
ERRCODE_INVALID_VALUE = -6
ERRCODE_NON_IMPLEMENTED = -7
ERRCODE_UNHANDLED_ERROR = -255


BODY_TYPE_COUPLED = POINT_TYPE_COUPLED = -1
BODY_TYPE_FREE = POINT_TYPE_FREE = 0
BODY_TYPE_FIXED = POINT_TYPE_FIXED = 1
ROD_TYPE_COUPLED = -2
ROD_TYPE_CPLDPIN = -1
ROD_TYPE_FREE = 0
ROD_TYPE_PINNED = 1
ROD_TYPE_FIXED = 2


ENDPOINT_A = ENDPOINT_BOTTOM = 0
ENDPOINT_B = ENDPOINT_TOP = 1


#                                  MoorDyn2.h
#  =============================================================================

def Create(filepath=""):
    """Creates the MoorDyn system, loading the MoorDyn input file

    Keyword arguments:
    filepath (str): The file to load. If "" is passed, then "Mooring/lines.txt"
                    will be considered

    Returns:
    cmoordyn.MoorDyn: The MoorDyn instance
    """
    import cmoordyn
    return cmoordyn.create(filepath)


def NCoupledDOF(instance):
    """Get the number of coupled Degrees Of Freedom (DOF)

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of DOFs
    """
    import cmoordyn
    return cmoordyn.n_coupled_dof(instance)


def SetVerbosity(instance, verbosity):
    """Set the verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    verbosity (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_verbosity(instance, verbosity)


def SetLogFile(instance, filepath):
    """Set the verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The log file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_logfile(instance, filepath)


def SetLogLevel(instance, verbosity):
    """Set the log file verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    verbosity (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.set_loglevel(instance, verbosity)


def Log(instance, level, msg):
    """Set the log file verbosity level of the MoorDyn instance

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    level (int): The Verbosity level. One of:
                     - moordyn.LEVEL_NONE
                     - moordyn.LEVEL_ERR
                     - moordyn.LEVEL_WRN
                     - moordyn.LEVEL_MSG
                     - moordyn.LEVEL_DBG
    msg (str): The message to log

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    import inspect

    head = "DBG "
    if level >= LEVEL_ERR:
        head = "ERR "
    elif level == LEVEL_WRN:
        head = "WRN "
    elif level == LEVEL_MSG:
        head = "MSG "
    caller = inspect.stack()[1]
    head = head + "{}:{} {}() : ".format(caller.filename,
                                         caller.lineno,
                                         caller.function)
    return cmoordyn.log(instance, level, head + msg)


def Init(instance, x, v):
    """Initializes the MoorDyn sytem, calculating initial conditions based on
    point positions specified by array x and their velocities in array v

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Keyword arguments:
    x (list): Position of the coupled points
    v (list): Velocity of the coupled points

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.init(instance, list(x), list(v), 0)


def Init_NoIC(instance, x, v):
    """Initializes the MoorDyn sytem, without calculating initial conditions

    Use this when you are planning to call moordyn.Load() afterwards

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Keyword arguments:
    x (list): Position of the coupled points
    v (list): Velocity of the coupled points

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.init(instance, list(x), list(v), 1)


def Step(instance, x, v, t, dt):
    """Compute a time step

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    x (list): Position of the coupled points
    v (list): Velocity of the coupled points
    t (float): The time instant
    dt (float): The time step

    Returns:
    list: The forces acting on the coupled points
    """
    import cmoordyn
    return cmoordyn.step(instance, list(x), list(v), t, dt)


def Close(instance):
    """This function deallocates the variables used by MoorDyn

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.close(instance)


def GetWaves(instance):
    """Get the wave kinematics instance

    The wave kinematics instance is only useful if WaveKin option is set to >= 2
    in the input file.

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    cmoordyn.MoorDynWaves: The waves manager
    """
    import cmoordyn
    return cmoordyn.get_waves(instance)


def GetSeafloor(instance):
    """Get the 3D seafloor instance

    The is not None just if a SeafloorPath option was set in the input file.

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    cmoordyn.MoorDynSeafloor: The seafloor instance
    """
    import cmoordyn
    try:
        return cmoordyn.get_seafloor(instance)
    except RuntimeError:
        return None


def ExternalWaveKinInit(instance):
    """Initializes the external Wave kinetics

    This is useless unless WaveKin option is set to 1 in the input file. If
    that is the case, remember calling this function after moordyn.Init()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.ext_wave_init(instance)


def ExternalWaveKinGetN(instance):
    """Get the number of points where the waves kinematics shall be provided

    This is useless unless WaveKin option is set to 1 in the input file.

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of points
    """
    import cmoordyn
    return cmoordyn.ext_wave_init(instance)


def ExternalWaveKinGetCoordinates(instance):
    """Get the points where the waves kinematics shall be provided

    The kinematics on those points shall be provided just if WaveKin is set
    to 1 in the input file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    list (n, 3): The list of points
    """
    import cmoordyn

    coords = cmoordyn.ext_wave_coords(instance)
    n = len(coords) // 3
    points = []
    for i in range(n):
        points.append([coords[3 * i], coords[3 * i + 1], coords[3 * i + 2]])
    return points


def GetWaveKinCoordinates(instance):
    """Just an alias for moordyn.ExternalWaveKinGetCoordinates()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    list (n, 3): The list of points
    """
    return ExternalWaveKinGetCoordinates(instance)


def ExternalWaveKinSet(instance, u, a, t):
    """Set the kinematics of the waves

    Use this function if WaveKin option is set to 1 in the input file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    u (list (n, 3)): The list of velocities evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    a (list (n, 3)): The list of accelerations evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    t (float): The simulation time

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn

    assert len(u) == len(a)
    uu, aa = [], []
    for i in range(len(u)):
        assert len(u[i]) == 3 and len(a[i]) == 3
        uu += u[i]
        aa += a[i]
    return cmoordyn.ext_wave_set(instance, uu, aa, t)


def SetWaveKin(instance, u, a, t):
    """Just an alias for moordyn.ExternalWaveKinSet()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    u (list (n, 3)): The list of velocities evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    a (list (n, 3)): The list of accelerations evaluated in the points provided
                     by moordyn.GetWaveKinCoordinates()
    t (float): The simulation time

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    return ExternalWaveKinSet(instance, uu, aa, t)


def GetNumberBodies(instance):
    """Get the number of bodies

    Remember that the first body index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of bodies
    """
    import cmoordyn
    return cmoordyn.get_number_bodies(instance)


def GetBody(instance, body):
    """ Get a rigid body

    Remember that the first body index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    body (int): The rigid body index, starting at 1

    Returns:
    cmoordyn.MoorDynBody: The rigid body instance
    """
    import cmoordyn
    return cmoordyn.get_body(instance, body)


def GetNumberRods(instance):
    """Get the number rods

    Remember that the rod index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of rods
    """
    import cmoordyn
    return cmoordyn.get_number_rods(instance)


def GetRod(instance, rod):
    """ Get a rod

    Remember that the first rod index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    rod (int): The rod index, starting at 1

    Returns:
    cmoordyn.MoorDynRod: The rod instance
    """
    import cmoordyn
    return cmoordyn.get_rod(instance, rod)


def GetNumberPoints(instance):
    """Get the number of points

    Remember that the first point index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of points
    """
    import cmoordyn
    return cmoordyn.get_number_points(instance)


def GetPoint(instance, point):
    """ Get a point

    Remember that the first point index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    point (int): The point index, starting at 1

    Returns:
    cmoordyn.MoorDynPoint: The point instance
    """
    import cmoordyn
    return cmoordyn.get_point(instance, point)


def GetNumberLines(instance):
    """Get the number of mooring lines

    Remember that the first line index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    int: The number of mooring lines
    """
    import cmoordyn
    return cmoordyn.get_number_lines(instance)


def GetLine(instance, line):
    """ Get a mooring line

    Remember that the first line index is 1

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    line (int): The line index, starting at 1

    Returns:
    cmoordyn.MoorDynLine: The line instance
    """
    import cmoordyn
    return cmoordyn.get_line(instance, line)


def GetFASTtens(instance, n_lines):
    """Get the horizontal and vertical components of the fairlead and anchor
    tensions

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    n_lines (int): The number of lines

    Returns:
    list: The horizontal fairlead tension at each line
    list: The vertical fairlead tension at each line
    list: The horizontal anchor tension at each line
    list: The vertical anchor tension at each line
    """
    import cmoordyn
    data = cmoordyn.get_fast_tens(instance, n_lines)
    return data[0], data[1], data[2], data[3]


def GetDt(instance):
    """Get the current model time step

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    dt (float): The time step
    """
    import cmoordyn
    cmoordyn.get_dt(instance)


def SetDt(instance, dt):
    """Set the model time step

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    dt (float): The new time step

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.set_dt(instance, dt)


def GetCFL(instance):
    """Get the current model Courant–Friedrichs–Lewy factor

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    cfl (float): The Courant–Friedrichs–Lewy factor
    """
    import cmoordyn
    cmoordyn.get_cfl(instance)


def SetCFL(instance, cfl):
    """Set the model Courant–Friedrichs–Lewy factor

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    cfl (float): The new Courant–Friedrichs–Lewy factor

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.set_cfl(instance, cfl)


def GetTimeScheme(instance):
    """Get the current time scheme name

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    name (str): The time scheme name
    """
    import cmoordyn
    cmoordyn.get_tscheme(instance)


def SetTimeScheme(instance, name):
    """Set the time scheme by its name

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    name (str): The new time scheme name

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.set_tscheme(instance, name)


def SaveState(instance, filepath):
    """Save the MoorDyn system state into a file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.save_state(instance, filepath)


def LoadState(instance, filepath):
    """Load and set a previously MoorDyn system state

    This can be used to load initial conditions exported with MoorPy

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.load_state(instance, filepath)


def Serialize(instance):
    """Serialize the MoorDyn system into a bytes array that can be restored
    afterwards to resume the simulation

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance

    Returns:
    Bytes array
    """
    import cmoordyn
    return cmoordyn.serialize(instance)


def Deserialize(instance, data):
    """Restores the MoorDyn system to a previous serialized state

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    data (bytes): The serialized data

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.deserialize(instance, data)


def Save(instance, filepath):
    """Save the MoorDyn system into a file that can be loaded afterwards to
    resume the simulation

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.save(instance, filepath)


def Load(instance, filepath):
    """Load a previously saved MoorDyn system, in order to resume the simulation

    You still must read the same definition file with moordyn.Create() and call
    moordyn.Init_NoIC()

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    None
    """
    import cmoordyn
    cmoordyn.load(instance, filepath)


def SaveVTK(instance, filename):
    """ Save the rod to a VTK (.vtu) file

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.save_vtk(instance, filename)


#                                  Waves.h
#  =============================================================================


def GetWavesKin(instance, x, y, z, seafloor=None):
    """ Get the wave kinematics

    Parameters:
    instance (cmoordyn.MoorDynWaves): The waves instance
    x (float): The x coordinate
    y (float): The y coordinate
    z (float): The z coordinate
    seafloor (cmoordyn.MoorDynSeafloor): The 3D seafloor instance

    Returns:
    u: The velocity (a tuple with 3 components)
    ud: The acceleration (a tuple with 3 components)
    zeta: The wave height
    pdyn: The dynamic pressure
    """
    import cmoordyn
    return cmoordyn.waves_getkin(instance, x, y, z, seafloor)


#                                  Seafloor.h
#  =============================================================================


def GetDepthAt(instance, x, y):
    """ Get the depth of the seafloor at some x and y

    Parameters:
    instance (cmoordyn.MoorDynSeafloor): The 3D seafloor instance
    x (float): The x coordinate
    y (float): The y coordinate

    Returns:
    depth: The output seafloor depth at that (x, y) point
    """
    import cmoordyn
    return cmoordyn.seafloor_getdepth(instance, x, y)
    

def GetAverageDepth(instance):
    """ Get the average of depth of the seafloor

    This value is calculated as the average value of every depth point

    If the rectilinear grid is not even, this average may not be the actual
    average depth of the surface the data describes.

    Parameters:
    instance (cmoordyn.MoorDynSeafloor): The 3D seafloor instance

    Returns:
    depth: The averaged seafloor depth
    """
    import cmoordyn
    return cmoordyn.seafloor_getavgdepth(instance)
    

def GetMinDepth(instance):
    """ The depth of the seafloor at the shallowest point

    Parameters:
    instance (cmoordyn.MoorDynSeafloor): The 3D seafloor instance

    Returns:
    depth: The minimum seafloor depth
    """
    import cmoordyn
    return cmoordyn.seafloor_getmindepth(instance)

#                                  Body.h
#  =============================================================================


def GetBodyID(instance):
    """ Get the body id

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    i: The body id
    """
    import cmoordyn
    return cmoordyn.body_get_id(instance)


def GetBodyType(instance):
    """ Get the body type

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    t: The body type
    """
    import cmoordyn
    return cmoordyn.body_get_type(instance)


def GetBodyState(instance):
    """ Get the body state

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The position (a tuple with 6 components)
    u: The velocity (a tuple with 6 components)
    """
    import cmoordyn
    return cmoordyn.body_get_state(instance)


def GetBodyPos(instance):
    """ Get the body position

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The position (a tuple with 3 components)
    """
    import cmoordyn
    return cmoordyn.body_get_pos(instance)


def GetBodyAngle(instance):
    """ Get the body Euler XYZ angles

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The angles (a tuple with 3 components)
    """
    import cmoordyn
    return cmoordyn.body_get_angle(instance)


def GetBodyVel(instance):
    """ Get the body velocity

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The velocity (a tuple with 3 components)
    """
    import cmoordyn
    return cmoordyn.body_get_vel(instance)


def GetBodyAngVel(instance):
    """ Get the body angular velocity

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    r: The angular velocity (a tuple with 3 components)
    """
    import cmoordyn
    return cmoordyn.body_get_angvel(instance)


def GetBodyForce(instance):
    """ Get the net force on the body

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    f: The force and the moment (a tuple with 6 components)
    """
    import cmoordyn
    return cmoordyn.body_get_force(instance)


def GetBodyM(instance):
    """ Get the mass matrix of the body

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance

    Returns:
    m: The mass matrix (a 6x6 components tuple)
    """
    import cmoordyn
    return cmoordyn.body_get_m(instance)


def SaveBodyVTK(instance, filename):
    """ Save the body to a VTK (.vtu) file

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.body_save_vtk(instance, filename)


def UseBodyVTK(instance, filename):
    """ Load the body representation from a 3d model file

    Parameters:
    instance (cmoordyn.MoorDynBody): The body instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.use_save_vtk(instance, filename)

#                                  Rod.h
#  =============================================================================


def GetRodID(instance):
    """ Get the rod id

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    i: The rod id
    """
    import cmoordyn
    return cmoordyn.rod_get_id(instance)


def GetRodType(instance):
    """ Get the rod type

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    i: The rod type
    """
    import cmoordyn
    return cmoordyn.rod_get_type(instance)


def GetRodForce(instance):
    """ Get the net force and moment acting over the rod

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    f: The rod force (a 6 components tuple)
    """
    import cmoordyn
    return cmoordyn.rod_get_force(instance)


def GetRodM(instance):
    """ Get the rod mass matrix

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    i: The rod mass matrix (a 6x6 components tuple)
    """
    import cmoordyn
    return cmoordyn.rod_get_m(instance)


def GetRodN(instance):
    """ Get the rod number of segments

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    n: The number of segments
    """
    import cmoordyn
    return cmoordyn.rod_get_n(instance)


def GetRodNumberNodes(instance):
    """ Get the rod number of nodes

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance

    Returns:
    n: The number of nodes
    """
    return GetRodN(instance) + 1


def GetRodNodePos(instance, i):
    """ Get a rod node position

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance
    i (int): The node index

    Returns:
    r: The node position
    """
    import cmoordyn
    return cmoordyn.rod_get_node_pos(instance, i)


def GetRodNodeVel(instance, i):
    """ Get a rod node position

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance
    i (int): The node index

    Returns:
    r: The node velocity
    """
    import cmoordyn
    return cmoordyn.rod_get_node_vel(instance, i)


def SaveRodVTK(instance, filename):
    """ Save the rod to a VTK (.vtu) file

    Parameters:
    instance (cmoordyn.MoorDynRod): The rod instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.rod_save_vtk(instance, filename)

#                                Point.h
#  =============================================================================


def GetPointID(instance):
    """ Get the point id

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    i: The point id
    """
    import cmoordyn
    return cmoordyn.point_get_id(instance)


def GetPointType(instance):
    """ Get the point type

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    t: The point type
    """
    import cmoordyn
    return cmoordyn.point_get_type(instance)


def GetPointPos(instance):
    """ Get the point position

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    v: The point position
    """
    import cmoordyn
    return cmoordyn.point_get_pos(instance)


def GetPointVel(instance):
    """ Get the point velocity

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    v: The point velocity
    """
    import cmoordyn
    return cmoordyn.point_get_vel(instance)


def GetPointForce(instance):
    """ Get the point force

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    f: The point force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.point_get_force(instance)


def GetPointM(instance):
    """ Get the point mass matrix

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    m: The mass matrix (a 3x3 components tuple)
    """
    import cmoordyn
    return cmoordyn.point_get_m(instance)


def GetPointNAttached(instance):
    """ Get the number of attached lines

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance

    Returns:
    n: The number of attached lines
    """
    import cmoordyn
    return cmoordyn.point_get_nattached(instance)


def GetPointAttached(instance, i):
    """ Get an attached line

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance
    i (int): The index of the attachment

    Returns:
    l: The attached line
    e: The endpoint
    """
    import cmoordyn
    return cmoordyn.point_get_attached(instance, i)


def SavePointVTK(instance, filename):
    """ Save the point to a VTK (.vtu) file

    Parameters:
    instance (cmoordyn.MoorDynPoint): The point instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.point_save_vtk(instance, filename)

#                                  Line.h
#  =============================================================================


def GetLineID(instance):
    """ Get the line id

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    i: The line id
    """
    import cmoordyn
    return cmoordyn.line_get_id(instance)


def GetLineN(instance):
    """ Get the line number of segments

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    n: The number of segments
    """
    import cmoordyn
    return cmoordyn.line_get_n(instance)


def GetLineNumberNodes(instance):
    """ Get the line number of nodes

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    n: The number of nodes
    """
    return GetLineN(instance) + 1


def GetLineUnstretchedLength(instance):
    """ Get the line unstretched length

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    l: The unstreched length
    """
    import cmoordyn
    return cmoordyn.line_get_ulen(instance)


def SetLineUnstretchedLength(instance, l):
    """ Set the line unstretched length

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    l (float): The unstreched length
    """
    import cmoordyn
    return cmoordyn.line_set_ulen(instance, l)


def SetLineUnstretchedLengthVel(instance, v):
    """ Set the line unstretched length rate of change

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    v (float): The unstreched length rate of change
    """
    import cmoordyn
    return cmoordyn.line_set_ulenv(instance, v)


def IsLineConstantEA(instance):
    """ Get whether the line is governed by a non-linear stiffness or a
    constant one

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    b: True if the stiffness of the line is constant, False if a non-linear
       stiffness has been set
    """
    import cmoordyn
    return cmoordyn.line_is_const_ea(instance) != 0


def GetLineConstantEA(instance):
    """ Get the constant stiffness of the line
    
    This value is useless if non-linear stiffness is considered
    
    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    ea: The constant stiffness EA value
    """
    import cmoordyn
    return cmoordyn.line_get_const_ea(instance)


def SetLineConstantEA(instance, ea):
    """ Set the constant stiffness of the line
    
    This value is useless if non-linear stiffness is considered
    
    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    ea (float): The constant stiffness EA value
    """
    import cmoordyn
    return cmoordyn.line_set_const_ea(instance, ea)


def IsLinePressBend(instance):
    """ Get whether the line pressure bending is considered or not
    
    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    b: True if the pressure bending of the line is enabled, False otherwise
    """
    import cmoordyn
    return cmoordyn.line_is_pbend(b)


def SetLinePressBend(instance, b):
    """ Set whether the line pressure bending is considered or not
    
    This value is useless if non-linear stiffness is considered
    
    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    b (bool): True if the pressure bending of the line shall be considered,
              False otherwise
    """
    import cmoordyn
    return cmoordyn.line_set_pbend(instance, b)


def SetLinePressInt(instance, p):
    """Set the line internal pressure values at the nodes

    Parameters:
    instance (cmoordyn.MoorDyn): The MoorDyn instance
    p (list): Pressure values
    """
    import cmoordyn
    return cmoordyn.line_set_pint(instance, list(p))


def GetLineNodePos(instance, i):
    """ Get a line node position

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    r: The node position
    """
    import cmoordyn
    return cmoordyn.line_get_node_pos(instance, i)


def GetLineNodeVel(instance, i):
    """ Get a line node velocity

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    rd: The node velocity
    """
    import cmoordyn
    return cmoordyn.line_get_node_vel(instance, i)


def GetLineNodeForce(instance, i):
    """ Get a line node net force

    To get the components of the force use moordyn.GetLineNodeTen() ,
    moordyn.GetLineNodeBendStiff(), moordyn.GetLineNodeWeight() ,
    moordyn.GetLineNodeDrag() , moordyn.GetLineNodeFroudeKrilov() and
    moordyn.GetLineNodeSeaBedForce()

    Note that the net force is *NOT* the sum of all those components. For
    instance, the tension contribution on the internal nodes is the difference
    between the tensions of the adjacent segments, while
    moordyn.GetLineNodeTen() is returning the averaged value.

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_force(instance, i)


def GetLineNodeTen(instance, i):
    """ Get a line node tension

    This comprises the axial stiffness as well as the internal damping

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    t: The node tension (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_ten(instance, i)


def GetLineNodeBendStiff(instance, i):
    """ Get a line node bending stiffness force

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_bend(instance, i)


def GetLineNodeWeight(instance, i):
    """ Get a line node weight and bouyancy

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_w(instance, i)


def GetLineNodeDrag(instance, i):
    """ Get a line node drag force

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_drag(instance, i)


def GetLineNodeFroudeKrilov(instance, i):
    """ Get a line node Froude-Krilov force

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_froudekrylov(instance, i)


def GetLineNodeSeabedForce(instance, i):
    """ Get a line node seabed reaction

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    f: The node force (a 3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_seabed(instance, i)


def GetLineNodeCurv(instance, i):
    """ Get a line node curvature

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    r: The node curvature
    """
    import cmoordyn
    return cmoordyn.line_get_node_curv(instance, i)


def GetLineNodeM(instance, i):
    """ Get a line node mass matrix

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    i (int): The node index

    Returns:
    r: The node mass matrix (a 3x3 components tuple)
    """
    import cmoordyn
    return cmoordyn.line_get_node_m(instance, i)


def GetLineFairTen(instance):
    """ Get the tension magnitude at the fairlead of a line

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    t: The tension magnitude
    """
    import cmoordyn
    return cmoordyn.line_get_fairlead_tension(instance)


def GetLineMaxTen(instance):
    """ Get the line maximum tension magnitude

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance

    Returns:
    t: The tension magnitude
    """
    import cmoordyn
    return cmoordyn.line_get_max_tension(instance)


def SaveLineVTK(instance, filename):
    """ Save the line to a VTK (.vtu) file

    Parameters:
    instance (cmoordyn.MoorDynLine): The line instance
    filepath (str): The file path

    Returns:
    int: 0 uppon success, an error code otherwise
    """
    import cmoordyn
    return cmoordyn.line_save_vtk(instance, filename)



"""
Copyright (c) 2022, Jose Luis Cercos-Pita <jlc@core-marine.com>

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution.

3. Neither the name of the copyright holder nor the names of its
   contributors may be used to endorse or promote products derived from
   this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

# This file is used to build self-contained Python wrappers, which does not
# depends on the prebuilt libmoordyn.so.2. This is ideal to be used within
# cibuildwheel, which is producing the artifacts that can be uploaded later to
# Pipy
# In general you do not want to use this script by your own. It is meant for the
# core developers to setup the pipy packages

import os
import shutil
from setuptools import setup, find_packages, Extension
import platform


# Collect the source code files
def get_sources(folder, sources=[]):
    for f in os.listdir(folder):
        if os.path.isdir(os.path.join(folder, f)):
            sources += get_sources(os.path.join(folder, f), sources=[])
        if not os.path.isfile(os.path.join(folder, f)):
            continue
        if not f.lower().endswith(".c") and not f.lower().endswith(".cpp"):
            continue
        sources.append(os.path.join(folder, f))
    return sources


# Check for needed files
def arefiles(files):
    return all([os.path.isfile(f) for f in files])

MODULE_PATH = os.path.join('wrappers', 'python', 'moordyn')
MOORDYN_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)),
                            'source')
MOORDYN_SRCS = [os.path.join('wrappers', 'python', 'cmoordyn.cpp')]

if os.path.isdir(MODULE_PATH):
    # We better copy the moordyn module on the root
    shutil.rmtree('moordyn', ignore_errors=True)
    shutil.copytree(MODULE_PATH, 'moordyn')

# Eigen needs at least C++ 14, and Moordyn itself uses C++ 17
extra_compile_args = ["-std=c++17"]
if platform.system() == "Windows":
    extra_compile_args.append("/std:c++17")
elif platform.system() == "Darwin":
    # To avoid errors with std::filesystem::path
    extra_compile_args.append("-mmacosx-version-min=10.15")
definitions = [('MoorDyn_EXPORTS', '1'),
               ('MOORDYN_MAJOR_VERSION', '2'),
               ('MOORDYN_MINOR_VERSION', '6'),
               ('MOORDYN_PATCH_VERSION', '0')]
include_dirs = [MOORDYN_PATH, 'install/include/moordyn']
library_dirs = ['install/lib/', 'install/lib64/']
libraries = ['moordyn']

cmoordyn = Extension('cmoordyn',
                     sources=MOORDYN_SRCS,
                     language='c++',
                     define_macros=definitions,
                     include_dirs=include_dirs,
                     library_dirs=library_dirs,
                     libraries=libraries,
                     extra_compile_args=extra_compile_args,
                     )

setup(
    packages=find_packages(include=['moordyn', 'moordyn.*']),
    ext_modules=[cmoordyn],
    package_data={"source": ["*.hpp", "*.h"],
                  "source.Util": ["*.hpp", "*.h"],
                  "source.Waves": ["*.hpp", "*.h"],
                  "source.Eigen": ["*"],
                  "source.Eigen.src": ["*"],
                  "source.Eigen.src.MetisSupport": ["*"],
                  "source.Eigen.src.SPQRSupport": ["*"],
                  "source.Eigen.src.AccelerateSupport": ["*"],
                  "source.Eigen.src.PaStiXSupport": ["*"],
                  "source.Eigen.src.IterativeLinearSolvers": ["*"],
                  "source.Eigen.src.SparseCholesky": ["*"],
                  "source.Eigen.src.misc": ["*"],
                  "source.Eigen.src.Core": ["*"],
                  "source.Eigen.src.Core.arch": ["*"],
                  "source.Eigen.src.Core.arch.GPU": ["*"],
                  "source.Eigen.src.Core.arch.AVX": ["*"],
                  "source.Eigen.src.Core.arch.NEON": ["*"],
                  "source.Eigen.src.Core.arch.MSA": ["*"],
                  "source.Eigen.src.Core.arch.SYCL": ["*"],
                  "source.Eigen.src.Core.arch.AltiVec": ["*"],
                  "source.Eigen.src.Core.arch.AVX512": ["*"],
                  "source.Eigen.src.Core.arch.SVE": ["*"],
                  "source.Eigen.src.Core.arch.SSE": ["*"],
                  "source.Eigen.src.Core.arch.HIP": ["*"],
                  "source.Eigen.src.Core.arch.HIP.hcc": ["*"],
                  "source.Eigen.src.Core.arch.ZVector": ["*"],
                  "source.Eigen.src.Core.arch.Default": ["*"],
                  "source.Eigen.src.Core.products": ["*"],
                  "source.Eigen.src.Core.functors": ["*"],
                  "source.Eigen.src.Core.util": ["*"],
                  "source.Eigen.src.Eigenvalues": ["*"],
                  "source.Eigen.src.SuperLUSupport": ["*"],
                  "source.Eigen.src.PardisoSupport": ["*"],
                  "source.Eigen.src.CholmodSupport": ["*"],
                  "source.Eigen.src.LU": ["*"],
                  "source.Eigen.src.LU.arch": ["*"],
                  "source.Eigen.src.UmfPackSupport": ["*"],
                  "source.Eigen.src.QR": ["*"],
                  "source.Eigen.src.Jacobi": ["*"],
                  "source.Eigen.src.SVD": ["*"],
                  "source.Eigen.src.StlSupport": ["*"],
                  "source.Eigen.src.Householder": ["*"],
                  "source.Eigen.src.Cholesky": ["*"],
                  "source.Eigen.src.SparseQR": ["*"],
                  "source.Eigen.src.SparseCore": ["*"],
                  "source.Eigen.src.Geometry": ["*"],
                  "source.Eigen.src.Geometry.arch": ["*"],
                  "source.Eigen.src.SparseLU": ["*"],
                  "source.Eigen.src.plugins": ["*"],
                  "source.Eigen.src.OrderingMethods": ["*"],
                  "source.Eigen.src.KLUSupport": ["*"],}
)



import moordyn
import numpy as np

### OPTIONAL ###
# import matplotlib.pyplot as plt
# import moorpy 

'''
This is a very simple example of how to use the MoorDyn Python interface. Users need to provide a moordyn input file before running.
This code is ready to run if a user provides an input file with a single coupled body. A more thorough explanation of how to drive 
MoorDyn can be found in `MoorDyn_standalone_demo.ipynb`.

The framework here is set up for a single coupled body. For other configurations, the state (x) and state derivatives (xd) need to be 
changed accordingly.
'''

def sin (period = 150, A = 10, axis = 0, dof = 6, x_initial = np.array([0, 0, 0, 0, 0, 0]), vector_size = 6, time = np.array([0])):
    '''
    A function that provides time series of states and state variables for simulating sinusoidal motion of a 6 DOF body in one of 
    the DOFs.

    Inputs
    ------
    `period` : `float`
        The period of the sinusoidal motion.
    `A` : `float`
        The amplitude of the sinusoidal motion.
    `axis` : `int`
        The axis along which the motion occurs (0 -> x, 1 -> y, 2 -> z, 3 -> roll, 4 -> pitch, 5 -> yaw).
    `dof` : `int`
        The degrees of freedom of the system.
    `x_initial` : `np.array`
        The initial state of the system.
    `vector_size` : `int`
        The size of the state vector. This should be equal to `DOF * the number of objects`. For example, 3 coupled points would have `DOF = 3` and `vector_size = 9`.
    `time` : `np.array`
        The time array for the simulation.

    Outputs
    -------
    `xp` : `np.array`
        The position time series of the system.
    `xdp` : `np.array`
        The velocity time series of the system.
    '''

    # axis 0 -> x, 1 -> y, 2 -> z, 3 -> roll, 4 -> pitch, 5 -> yaw
    xp = np.zeros((len(time),6))
    
    # Time info
    dt = time[1] - time[0]

    # Wave properties
    T = period / dt
    omega = (2*np.pi)/T
    
    for i in range(len(time)):
        xp[i,axis] = A * np.sin(i*omega)

    xdp = np.zeros((len(time),6))
    xold = np.zeros(6)
    # calculate velocities using finite difference
    for i in range(len(time)):
        xdp [i] = (xp[i] - xold)/dt
        xold =  xp[i]
    for i in range(len(time)):
        if i == 0:
            x[i,:] = x_initial
        else:
            j = 0
            while j < vector_size:
                x[i,j:j+dof] = x[i-1,j:j+dof] + xdp[i, 0:dof] * dt
                xd[i,j:j+dof] = xdp[i, 0:dof]
                j += dof

    return x, xd

# Create time array
tMax = 0.5 # max time for running time series
dt = 0.001 # coupling timestep
time = np.arange(0, tMax, dt) # time series

# Build arrays of state and state derivatives
size = (len(time), 6)
x = np.zeros(size)
xd = np.zeros(size)

# Create the MoorDyn system from the input file. 
# NOTE TO USERS:
#   Change this to your own input file. 
#   It must have one coupled body and no other coupled objects. 
system = moordyn.Create("<file name>")

# You can get the initial positions and velocities from the system itself using the MoorDyn-C API
body = moordyn.GetBody(system, 1)
state = moordyn.GetBodyState(body) # tuple with (x, xd)

# Set the first time step of the state and state derivatives to the initial values from MoorDyn
x[0,:] = np.array(state[0]) # x, y, z, roll, pitch, yaw
xd[0,:] = np.array(state[1]) # xdot, ydot, zdot, roll_dot, pitch_dot, yaw_dot

# Setup the initial condition
moordyn.Init(system, x[0,:], xd[0,:])

# Get the state and state derivative time series
x_t, xd_t = sin(period = 5, A = 1, axis = 0, dof = 6, x_initial = x[0,:], vector_size = 6, time = time)

# Run the simulation for five timesteps
for i in range(1, len(time)):
    # call the MoorDyn step function
    f = moordyn.Step(system, x_t[i,:], xd_t[i,:], time[i], dt)    # force value on coupled DOF returned here in array

# Alright, time to finish!
moordyn.Close(system)


### Optional ###

# # Plot and animate the results using MoorPy

# # Create a MoorPy system (this loads the moordyn outputs found in <dirname>/<rootname>_<object>#.out, where <object> is either Line or Rod and # is the corresponding number)
# ms = moorpy.System(file=<path+rootname+extension>, dirname=<path>, rootname=<rootname>, qs = 0, Fortran = False) # qs tells MoorPy it is loading a MoorDyn system, Fortran = False tells it to use the MoorDyn-C output format

# # Plot the results at t = 0 in red
# ms.plot(color="red", time = 0)

# # animate the results
# animation = ms.animateLines()

# plt.show()


import numpy as np

T = 15.0
A = 1.0
UX = 1.25
x = np.linspace(-450.0, 50.0, num=501)
y = np.linspace(-2.0, 2.0, num=5)
z = np.linspace(-51.0, 5.0, num=57)
t = np.linspace(0.0, 15.0, num=101)
zeta = A * np.sin(4.0 * 2.0 * np.pi * t / T)

with open('water_grid.txt', 'w') as f:
    f.write('--------------------- MoorDyn Waves grid File ----------------------------------\n')
    f.write('List of grid points, in 3 blocks (x, y, z)\n')
    f.write('Each block starts with a 2 (i.e. list of coords), and then th coordinates (m)\n')
    for l in (x, y, z):
        txt = '1\n'
        for p in l:
            txt = txt + str(p) + ' '
        f.write(txt.strip() + '\n')

with open('wave_elevation.txt', 'w') as f:
    for i, tt in enumerate(t):
        f.write(str(tt) + ' ' + str(zeta[i]) + '\n')

with open('current_profile.txt', 'w') as f:
    f.write('--------------------- MoorDyn steady currents File ----------------------------------\n')
    f.write('Tabulated file with the water currents components\n')
    f.write('z (m), ux (m/s), uy (m/s), uz (m/s)\n')
    for zz in z:
        ux = 0.0
        if zz > -50.0 and zz <= -10.0:
            ux = UX * (1.0 - (-zz - 10.0) / 40.0)
        elif zz > -10.0 and zz < -A:
            ux = UX * (1.0 - (zz + 10.0) / (10.0 - A))
        f.write(str(zz) + " " + str(ux) + " 0 0\n")



import sys
from unittest import TestCase, main as unittest_main
import os
import math
import moordyn


class ICTest(TestCase):
    def setUp(self):
        pass

    def test_no_input(self):
        try:
            system = moordyn.Create('file.invalid')
        except RuntimeError:
            return
        self.assertEqual(system, None, "The creation should fail")

    def test_one_iter(self):
        # 1st IC generation stage: The stationary solver
        system = moordyn.Create("Mooring/riser/riser_ic1.txt")
        point = moordyn.GetPoint(system, 4)
        r = moordyn.GetPointPos(point)
        v = [0, 0, 0]
        moordyn.Init(system, r, v)
        moordyn.SaveState(system, "Mooring/riser/riser.ic")
        moordyn.Close(system)
        # 2nd IC generation stage: The dynamic solver
        system = moordyn.Create("Mooring/riser/riser_ic2.txt")
        point = moordyn.GetPoint(system, 4)
        r = moordyn.GetPointPos(point)
        moordyn.Init(system, r, v)
        moordyn.SaveState(system, "Mooring/riser/riser.ic")
        moordyn.Close(system)
        # The simulation itself
        system = moordyn.Create("Mooring/riser/riser.txt")
        point = moordyn.GetPoint(system, 4)
        r = moordyn.GetPointPos(point)
        moordyn.Init(system, r, v)
        T = 10
        A = 0.0
        dt = T / 100
        t = 0.0
        step = 0
        while t <= 1 * T:
            rorg = moordyn.GetPointPos(point)
            rdst = list(rorg)
            rdst[2] = r[2] + A * math.sin(2 * math.pi * (t + dt) / T)
            v = [(rdst[i] - rorg[i]) / dt for i in range(len(rorg))]
            moordyn.Step(system, rorg, v, t, dt)
            try:
                moordyn.SaveVTK(system,
                                f"Mooring/riser/vtk/out.{step:05d}.vtm")
            except:
                pass
            t += dt
            step += 1
        moordyn.Close(system)


if __name__ == '__main__':
    unittest_main()



import sys
from unittest import TestCase, main as unittest_main
import os
import moordyn
from scipy.spatial.transform import Rotation as R
import numpy as np
from test_minimal import setup_case


ENUREF = [[1, 0, 0],
          [0, 1, 0],
          [0, 0, 1]]
ANGLE = 5


def rotate_vecs(vecs, angles, seq="XYZ", degrees=True):
    r = R.from_euler(seq, angles, degrees=degrees)
    return [r.apply(vec) for vec in vecs]


def compare_vecs(vecs, refs, tol=1e-5):
    assert len(vecs) == len(refs)
    for i in range(len(vecs)):
        assert len(vecs[i]) == len(refs[i])
        for j in range(len(vecs[i])):
            assert abs(vecs[i][j] - refs[i][j]) < 1e-5


def abseuler2releuler(org, dst, seq="XYZ", degrees=True):
    r0 = R.from_euler(seq, org, degrees=degrees).inv()
    r1 = R.from_euler(seq, dst, degrees=degrees)
    return (r1 * r0).as_euler(seq, degrees=degrees)
    

def rotate_moordyn(system, roll, pitch, yaw, t, dt=1.0, degrees=True):
    if degrees:
        roll = np.radians(roll)
        pitch = np.radians(pitch)
        yaw = np.radians(yaw)
    pts = [moordyn.GetPoint(system, i + 1) for i in range(3)]
    body = moordyn.GetBody(system, 1)
    r, _ = moordyn.GetBodyState(body)
    u = [0] * 3 + [roll / dt, pitch / dt, yaw / dt]
    moordyn.Step(system, r, u, t, dt)
    t += dt
    r = [moordyn.GetPointPos(pt) for pt in pts]
    return r, t


class RotationTests(TestCase):
    def setUp(self):
        pass

    def test_nwu(self):
        tmp_folder = setup_case("rotations/nwu.txt")
        cwd = os.getcwd()
        os.chdir(tmp_folder)
        system = moordyn.Create()
        os.chdir(cwd)
        pts = [moordyn.GetPoint(system, i + 1) for i in range(3)]
        r = [0] * 6
        u = [0] * 6
        moordyn.Init(system, r, u)
        # Check the intial condition
        compare_vecs([moordyn.GetPointPos(pt) for pt in pts], ENUREF)
        t = 0
        # Roll test
        r, t = rotate_moordyn(system, ANGLE, 0, 0, t)
        compare_vecs(r, rotate_vecs(ENUREF, ANGLE, seq='X'))
        assert(r[1][2] > 0.0)  # Portside up, i.e. starboard down
        r, t = rotate_moordyn(system, -ANGLE, 0, 0, t)
        compare_vecs(r, ENUREF)
        # Pitch test
        r, t = rotate_moordyn(system, 0, ANGLE, 0, t)
        compare_vecs(r, rotate_vecs(ENUREF, ANGLE, seq='Y'))
        assert(r[0][2] < 0.0)  # Bow down
        r, t = rotate_moordyn(system, 0, -ANGLE, 0, t)
        compare_vecs(r, ENUREF)
        # Yaw test
        r, t = rotate_moordyn(system, 0, 0, ANGLE, t)
        compare_vecs(r, rotate_vecs(ENUREF, ANGLE, seq='Z'))
        assert(r[0][1] > 0.0)  # Bow to portside
        r, t = rotate_moordyn(system, 0, 0, -ANGLE, t)
        compare_vecs(r, ENUREF)

        # Test the angles order. For that we are rotating it in all directions
        # simultaneously
        r, t = rotate_moordyn(system, ANGLE, ANGLE, ANGLE, t)
        compare_vecs(r, rotate_vecs(ENUREF, [ANGLE, ANGLE, ANGLE]))
        r, t = rotate_moordyn(system, -ANGLE, -ANGLE, -ANGLE, t)
        compare_vecs(r, ENUREF)

        # Test the rotations concatenations. To do that we are again rotating on
        # the 3 axis simultaneously. However, this time we rotate twice, so the
        # final angle shall be 2*ANGLE in all directions
        rotate_moordyn(system, ANGLE, ANGLE, ANGLE, t)
        r, t = rotate_moordyn(system, ANGLE, ANGLE, ANGLE, t)
        roll, pitch, yaw = abseuler2releuler([ANGLE, ANGLE, ANGLE],
                                            [2 * ANGLE, 2 * ANGLE, 2 * ANGLE])
        compare_vecs(r, rotate_vecs(rotate_vecs(ENUREF, [ANGLE, ANGLE, ANGLE]),
                                    [roll, pitch, yaw]))
        compare_vecs(r, rotate_vecs(ENUREF, [2 * ANGLE, 2 * ANGLE, 2 * ANGLE]))

        moordyn.Close(system)


if __name__ == '__main__':
    unittest_main()
